from multiprocessing import Pool as ThreadPool

import numpy as np
import multiprocessing as mp
from functools import partial
import math

def find_all_dist_with_target(matrix,
                              index_array,
                              total_indexes,
                              feature_weights,
                              target_dp_index):
    """
        Finds all distances between the target and the other points.
    """
    distances = np.zeros(total_indexes - (target_dp_index + 1))
    for index, dp_index in enumerate(index_array[
                                     target_dp_index + 1:]):
        distances[index] = weighted_eudis(matrix[target_dp_index],
                                          matrix[dp_index],
                                          feature_weights)
        # self.__pbar.update(1)
    # shortest_dp_index = np.argmin(distances)

    all_distances_to_target = dict()

    all_distances_to_target[target_dp_index] = distances

    return all_distances_to_target


def find_all_distances_in_matrix(matrix,
                                 index_array,
                                 total_indexes,
                                 feature_weights):
    pool = ThreadPool(mp.cpu_count()-2)

    func = partial(find_all_dist_with_target,
                   matrix,
                   index_array,
                   total_indexes,
                   feature_weights)
    all_dp_distances = list(
        pool.imap_unordered(func,
                            index_array[:-1]))

    # Store dps relationships and the distances
    all_dp_dist_list = [np.array([])] * matrix.shape[0]

    # Convert map to list
    for dp_dict in all_dp_distances:
        all_dp_dist_list[list(dp_dict.keys())[0]] = \
            list(dp_dict.values())[0]

    return all_dp_dist_list


def weighted_eudis(v1,
                   v2,
                   feature_weights):
    dist = [((a - b) ** 2) * w for a, b, w in zip(v1, v2,
                                                  feature_weights)]
    dist = math.sqrt(sum(dist))
    return dist

def fast_eudis(v1,
               v2):
    dist = [((a - b) ** 2) for a, b in zip(v1, v2)]
    dist = math.sqrt(sum(dist))
    return dist


from eflow._hidden.constants import SYS_CONSTANTS
from eflow.utils.sys_utils import write_object_text_to_file
from eflow.utils.misc_utils import string_condtional

import pandas as pd

import ipywidgets as widgets
from IPython.display import display
from ipywidgets import Layout
import json
import uuid
import os.path
import copy


class DataCleaningWidget():

    def __init__(self,
                 require_input=None,
                 data_cleaning_options=None):
        """
        df: pd.Dataframe
            Pandas dataframe object

        project_name: string
            Appending directory structure/name to the absolute path of the
            output directory.

        overwrite_full_path: string
            Define the entire output path of the cleaner.
        """

        # Throw errors here
        if data_cleaning_options is None:
            return

        if require_input is None:
            return

        # Written conditionals for functions requiring input fields
        self.__require_input = require_input
        self.__data_cleaning_options = data_cleaning_options

        # Define needed widgets
        self.__feature_cleaning_options_w = None
        self.__features_w = None
        self.__options_w = None
        self.__input_w = None
        self.__selected_options_w = None
        self.__full_widgets_ui = None

        # ---
        self.__df_features = None
        self.__json_file_path = None
        self.__feature_input_holder = dict()
        self.__feature_zscore_holder = dict()
        self.__selected_options = dict()

    def get_user_inputs(self):
        return self.__selected_options, \
               {feature: value for feature, value in
                self.__feature_input_holder.items() if value}, \
               {feature: value for feature, value in
                self.__feature_zscore_holder.items() if value}

    def run_widget(self,
                   nan_feature_names,
                   df_features):
        """

            Spin up widget display for in jupyter UI.

        Args:
            df: pd.Dataframe
                A pandas dataframe object

            df_features: DataFrameTypes
                DataFrameTypes object; organizes feature types into groups.

        Returns:
            Returns a UI widget to create a JSON file for cleaning.
        """
        if not nan_feature_names:
            print("Nan features must not be an empty list!")
            return None

        self.__df_features = df_features

        for feature_name in nan_feature_names:
            if feature_name not in self.__selected_options:
                self.__selected_options[feature_name] = "Ignore feature"
        # self.__selected_options = {feature_name: "Ignore feature"
        #                            for feature_name in nan_feature_names}

        feature_cleaning_options = {col_feature_name:self.__data_cleaning_options[
            self.__get_dtype_key(df_features,
                                 col_feature_name)].keys()
                                    for col_feature_name in nan_feature_names}

        self.__feature_cleaning_options_w = {key: widgets.Select(
            options=feature_cleaning_options[key],
            layout=Layout(width='50%',
                          height='300px'))
            for key in feature_cleaning_options}

        self.__features_w = widgets.Select(
            options=list(feature_cleaning_options.keys()),
            layout=Layout(width='50%',
                          height='175px')
        )

        self.__selected_options_w = widgets.Select(
            options=list(self.__format_selected_options()),
            layout=Layout(width='80%',
                          height='375px')
        )

        self.__init_update_updateable_widgets()


        self.__input_w = widgets.Text(
            value='',
            placeholder='Replace Value',
            description='Input:',
            disabled=False,
            visible=False,
            layout=Layout(left='49.06%',
                          bottom='250px')
        )

        # ---
        self.__zscore_w = widgets.Text(
            value='',
            placeholder='Z-Score Value',
            description='Z Score:',
            disabled=False,
            visible=False,
            layout=Layout(left='49%',
                          bottom="330px", )
        )

        # Link functions with non-updateable widgets
        self.__features_w.observe(self.__select_feature,
                                  'value')
        self.__features_w.observe(self.__hide_init_zscore_w,
                                  'value')

        self.__input_visible = False
        self.__zscore_visible = False

        self.__zscore_w.observe(self.__validate_save_zscore)
        self.__input_w.observe(self.__validate_save_input)
        self.__options_w.observe(self.__set_input_value,
                                 "value")
        self.__features_w.observe(self.__set_zscore_value,
                                  "value")

        self.__validate_save_zscore(None)
        self.__validate_save_input(None)

        # Setup and display full UI
        self.__full_widgets_ui = widgets.interactive(
            self.__save_option,
            Features=self.__features_w,
            Options=self.__options_w,
            Text_Input=self.__input_w,
            Z_Score_Input=self.__zscore_w,
            Selected=self.__selected_options_w
        )

        display(self.__full_widgets_ui)

    def __init_update_updateable_widgets(self):
        init = self.__features_w.value
        self.__options_w = self.__feature_cleaning_options_w[init]

        if self.__input_w:
            self.__hide_init_input_w(None)

    def __validate_save_input(self,
                              _):
        """
        Returns:
            Ensures the input field is within specified parameters defined
            by the 'require_input' dictionary.
        """
        #
        # if len(self.__input_w.value) == 0:
        #     return
        #
        # if self.__features_w.value in self.__feature_input_holder and self.__feature_input_holder[self.__features_w.value]:
        #     self.__feature_input_holder.pop(self.__features_w.value,
        #                                     None)

        if self.__options_w.value in self.__require_input:

            feature_type = self.__get_dtype_key(
                    self.__df_features,
                    self.__features_w.value)
            if len(self.__input_w.value) > 0:
                self.__input_w.value = ''.join(
                    [i for i in self.__input_w.value if i.isdigit() or i == '.'
                     or i == "-"])

                if self.__require_input[self.__options_w.value] is not None \
                        and not string_condtional(self.__input_w.value,
                                                  self.__require_input[
                                                      self.__options_w.value]) and \
                        (self.__input_w.value.strip() != "-" and self.__input_w.value.strip() != "."):
                    self.__input_w.value = self.__input_w.value[:-1]

            self.__feature_input_holder[self.__features_w.value] = \
                self.__input_w.value
        else:
            self.__feature_input_holder[self.__features_w.value] = ""
            self.__input_w.value = ""

    def __validate_save_zscore(self,
                               _):
        """
        Returns:
            Validates the z-score widget and saves the value of the z-score
            with selected feature.
        """

        self.__zscore_w.value = "".join(x for x in str(
            self.__zscore_w.value) if x.isdigit() or x == ".")

        if self.__zscore_w.value == ".":
            self.__zscore_w.value = "0.0"

        if self.__zscore_w.value != "":
            # if self.__zscore_w.value[-1] == ".":
            #     self.__zscore_w.value += "0"

            self.__feature_zscore_holder[self.__features_w.value] = \
                self.__zscore_w.value

    def __hide_init_input_w(self,
                           _):

        if self.__options_w.value in self.__require_input:
            self.__input_w.layout.visibility = 'visible'
            self.__input_visible = True
        else:
            self.__input_w.layout.visibility = 'hidden'
            self.__input_visible = False

    def __hide_init_zscore_w(self,
                             _):
        if self.__get_dtype_key(self.__df_features,
                                self.__features_w.value) == "Number":
            self.__zscore_w.layout.visibility = "visible"
            self.__zscore_visible = True
        else:
            self.__zscore_w.layout.visibility = "hidden"
            self.__zscore_visible = False

    def __save_option(self,
                      **func_kwargs):
        if not self.__selected_options:
            return None

        self.__selected_options[func_kwargs["Features"]] = self.__options_w.value
        self.__selected_options_w.options = self.__format_selected_options()

        # print(func_kwargs["Features"])
        # print("\n\n\t     Feature option Review\n\t   " + "*" * 25)
        # for feature, option in self.__selected_options.items():
        #
        #     if option[0:3] == "---":
        #         option = "Ignore Feature"
        #
        #     print("\n\t   Feature: {0}\n\t   Option:  {1}\n".format(
        #         feature,
        #         option)
        #           + "           " + "----" * 7)
        self.__hide_init_input_w(None)

    def __set_input_value(self,
                          _):

        if self.__input_visible:
            if self.__features_w.value in self.__feature_input_holder:
                self.__input_w.value = self.__feature_input_holder[self.__features_w.value]
            else:
                self.__input_w.value = ""
        else:
            self.__input_w.value = ""
            if self.__features_w.value in self.__feature_input_holder:
                self.__feature_input_holder[self.__features_w.value] = ""


    def __set_zscore_value(self,
                          _):

        if self.__zscore_visible:
            if self.__features_w.value in self.__feature_zscore_holder:
                self.__zscore_w.value = self.__feature_zscore_holder[self.__features_w.value]
            else:
                self.__zscore_w.value = ""
        else:
            self.__zscore_w.value = ""
            if self.__features_w.value in self.__feature_zscore_holder:
                self.__feature_zscore_holder[self.__features_w.value] = ""


    def __select_feature(self,
                         feature):
        """
        Returns:
            When a feature selection is chosen all the widgets are
            re-initialized.
        """

        write_object_text_to_file(self.__feature_input_holder,
                                  os.getcwd(),
        "fuckk")

        self.__init_update_updateable_widgets()

        new_i = widgets.interactive(self.__save_option,
                                    Features=self.__features_w,
                                    Options=self.__options_w,
                                    Text_Input=self.__input_w,
                                    Z_Score_Input=self.__zscore_w,
                                    Selected=self.__selected_options_w)

        self.__full_widgets_ui.children = new_i.children

        self.__set_input_value("really?")

    # --- General functionality

    def __format_selected_options(self):

        formated_list = []
        for feature, option in self.__selected_options.items():

            if option[0:3] in "---":
                option = "Ignore Feature"

            formated_string = "Feature:{:<20s} Option:{:s}".format(feature,
                                                                   option)

            tmp = str()
            for i, char in enumerate(formated_string):

                if i == len(formated_string)-1:
                    tmp += char
                    break

                if char == " " and formated_string[i - 1] != ":" and formated_string[i - 1] == " " and \
                        formated_string[i + 1] == " ":
                    tmp += "-"
                else:
                    tmp += char
            formated_string = tmp

            formated_list.append(formated_string)

        return formated_list

    def __get_dtype_key(self,
                        df_features,
                        col_feature_name):
        """
        Args:
            df_features:
                DataFrameTypes object; organizes feature types into groups.

            col_feature_name:
                Pandas column name.

        Returns:
            Returns back the data type of the feature that is created
        """

        if col_feature_name in df_features.continuous_numerical_features():
            return "Number"
        elif col_feature_name in df_features.categorical_features() or col_feature_name in df_features.string_features():
            return "Category"
        elif col_feature_name in df_features.bool_features():
            return "Bool"
        else:
            return "Unknown"


import warnings

__author__ = "Eric Cacciavillani"
__copyright__ = "Copyright 2019, eFlow"
__credits__ = ["Eric Cacciavillani"]
__license__ = "MIT"
__maintainer__ = "EricCacciavillani"
__email__ = "eric.cacciavillani@gmail.com"



def custom_formatwarning(msg,
                         *args,
                         **kwargs):
    # ignore everything except the message
    return str(msg) + '\n'

class EflowWarning(UserWarning):
    warnings.formatwarning = custom_formatwarning



from eflow._hidden.parent_objects import JupyterWidget

import ipywidgets as widgets
from IPython.display import display
from ipywidgets import Layout


class StringCleaningWidget(JupyterWidget):

    def __init__(self):

        JupyterWidget.__init__(self,
                               self.__class__.__name__)

        # Define needed widgets
        self.__feature_cleaning_options_w = None
        self.__features_w = None
        self.__options_w = None
        self.__input_w = None
        self.__full_widgets_ui = None

        # ---
        self.__require_input = dict()

    def run_widget(self,
                   df,
                   df_features):
        """


        Args:
            df: pd.Dataframe
                A pandas dataframe object

            df_features: DataFrameType
                DataFrameTypes object; organizes feature types into groups.

        Returns:
            Returns a UI widget to create a JSON file for cleaning.
        """
        self.__feature_option_dict = dict()

        self.__selected_options = {feature_name: "None"
                                   for feature_name in
                                   df_features.string_features()}

        feature_cleaning_options = {feature_name: ["None"]
                                    for feature_name in
                                    df_features.string_features()}

        self.__feature_options_w = {key: widgets.Select(
            options=feature_cleaning_options[key],
            layout=Layout(width='50%',
                          height='300px'))
            for key in feature_cleaning_options}

        self.__features_w = widgets.Select(
            options=list(feature_cleaning_options.keys()),
            layout=Layout(width='50%',
                          height='175px')
        )

        self.__update_widgets()

        # ---
        self.__file_name_w = widgets.Text(
            value='Default Data Cleaning',
            placeholder='Replace Value',
            description='File Name:',
            disabled=False,
            visible=False,
            layout=Layout(left='590px')
        )

        # ---
        self.__submit_button = widgets.Button(
            description='Create JSON File from options',
            color="#ff1122",
            layout=Layout(left='100px',
                          bottom="5px",
                          width='40%', ))

        self.__input_w = widgets.Text(
            value='',
            placeholder='Replace Value',
            description='Input:',
            disabled=False,
            visible=False,
            layout=Layout(width='50%')
        )

        # Link functions with non-updateable widgets
        self.__features_w.observe(self.__select_feature,
                                  'value')

        self.__input_w.observe(self.__validate_save_input_w)

        # Setup and display full UI
        self.__full_widgets_ui = widgets.interactive(
            self.__save_option,
            Features=self.__features_w,
            Options=self.__options_w,
            Text_Input=self.__input_w,
        )

        display(self.__full_widgets_ui)

    def __update_widgets(self):
        init = self.__features_w.value
        self.__options_w = self.__feature_options_w[init]

    def __validate_save_input_w(self,
                                _):
        """
        Returns:
            Ensures the input field is within specified parameters defined
            by the 'require_input' dictionary.
        """

        pass

    def __hide_init_input_w(self,
                            _):

        if self.__options_w.value in self.__require_input:
            self.__input_w.layout.visibility = 'visible'
        else:
            self.__input_w.layout.visibility = 'hidden'
            self.__input_w.layout.visibility = 'visible'
            # self.__input_w.value = ""

    def __save_option(self,
                      **func_kwargs):
        self.__selected_options[
            func_kwargs["Features"]] = self.__options_w.value
        self.__hide_init_input_w(None)

    def __select_feature(self,
                         feature):
        """
        Returns:
            When a feature selection is chosen all the widgets are
            re-initialized.
        """

        self.__update_widgets()

        new_i = widgets.interactive(self.__save_option,
                                    Features=self.__features_w,
                                    Options=self.__options_w,
                                    Text_Input=self.__input_w)

        self.__full_widgets_ui.children = new_i.children

# StringCleaningWidget().run_widget(df,
#                                   df_features)


# Getting Sklearn Models
from sklearn.decomposition import PCA
from sklearn.cluster import KMeans
from scipy.cluster.hierarchy import linkage, dendrogram
from sklearn.preprocessing import StandardScaler
from sklearn.cluster import AgglomerativeClustering
from sklearn.neighbors import kneighbors_graph

# Visuals libs
import matplotlib.pyplot as plt
import pylab as pl
import seaborn as sns
from IPython.display import display, HTML

# Misc
from scipy.stats import zscore
from kneed import DataGenerator, KneeLocator
import pandas as pd
import six
import sys
import numpy as np
import os
import copy

__author__ = "Eric Cacciavillani"
__copyright__ = "Copyright 2019, eFlow"
__credits__ = ["Eric Cacciavillani"]
__license__ = "MIT"
__maintainer__ = "EricCacciavillani"
__email__ = "eric.cacciavillani@gmail.com"


class ClusterMaster:

    def __init__(self,
                 df,
                 apply_pca=True,
                 pca_perc=.8,
                 project_name="Default",
                 overwrite_figure_path=None
                 ):

        if overwrite_figure_path:
            output_fig_sub_dir = overwrite_figure_path
        else:
            if pca_perc > 1:
                pca_perc = 1
            output_fig_sub_dir = "/Figures/" + project_name +\
                                 "/Clustering_PCA={0}".format(pca_perc)

        # Project directory structure
        self.__PROJECT = enum(
            PATH_TO_OUTPUT_FOLDER=''.join(
                os.getcwd().partition('/eflow')[0:1]) + output_fig_sub_dir)

        # Define model
        self.__all_cluster_models = dict()

        # --- Apply pca ---
        if apply_pca:

            # Create scaler object
            scaler = StandardScaler()
            scaled = scaler.fit_transform(df)

            print("\nInspecting scaled results!")
            self.__inspect_feature_matrix(matrix=scaled,
                                          feature_names=df.columns)

            pca, scaled = self.__visualize_pca_variance(scaled)

            # Generate "dummy" feature names
            pca_feature_names = ["PCA_Feature_" +
                                 str(i) for i in range(1,
                                                       len(df.columns) + 1)]

            print("\nInspecting applied pca results!")
            self.__inspect_feature_matrix(matrix=scaled,
                                          feature_names=pca_feature_names)

            if pca_perc < 1.0:
                # Find cut off point on cumulative sum
                cutoff_index = np.where(
                    pca.explained_variance_ratio_.cumsum() > pca_perc)[0][0]
            else:
                cutoff_index = scaled.shape[1] - 1

            print(
                "After applying pca with a cutoff percentage {0}%"
                " for the cumulative index. Using features 1 to {1}".format(
                    pca_perc, cutoff_index + 1))

            print("Old shape {0}".format(scaled.shape))

            scaled = scaled[:, :cutoff_index + 1]
            pca_feature_names = pca_feature_names[0: cutoff_index + 1]

            print("New shape {0}".format(scaled.shape))

            scaled = scaler.fit_transform(scaled)

            print("\nInspecting re-applied scaled results!")
            self.__inspect_feature_matrix(matrix=scaled,
                                          feature_names=pca_feature_names)

            self.__scaled = scaled

        # Assumed PCA has already been applied; pass as matrix
        else:
            self.__scaled = df.values

    def __display_rank_graph(self, feature_names, metric,
                             output_path, model_name,
                             title="", y_title="", x_title="",):
        """
            Darker colors have higher rankings (values)
        """
        plt.figure(figsize=(7, 7))

        # Init color ranking fo plot
        # Ref: http://tinyurl.com/ydgjtmty
        pal = sns.color_palette("GnBu_d", len(metric))
        rank = np.array(metric).argsort().argsort()
        ax = sns.barplot(y=feature_names, x=metric,
                         palette=np.array(pal[::-1])[rank])
        plt.xticks(rotation=0, fontsize=15)
        plt.yticks(fontsize=15)
        plt.xlabel(x_title, fontsize=20, labelpad=20)
        plt.ylabel(y_title, fontsize=20, labelpad=20)
        plt.title(title, fontsize=15)
        self.__image_processing_utils(output_path,
                              model_name + "_Cluster_Count")
        plt.show()
        plt.close()

    def visualize_hierarchical_clustering(self,
                                          linkage_methods=None):
        """
            Displays hierarchical cluster graphs with provided methods.
        """

        if not linkage_methods:
            linkage_methods = ["complete",
                               "single",
                               "weighted",
                               "ward",
                               "average",
                               "centroid",
                               "median"]

        for method in linkage_methods:

            plt.figure(figsize=(12, 7))
            # Calculate the linkage: mergings
            mergings = linkage(self.__scaled, method=method)

            # Plot the dendrogram, using varieties as labels
            dendrogram(mergings,
                       labels=list(range(0, len(self.__scaled,))),
                       leaf_rotation=90,
                       leaf_font_size=3)

            plt.title("hierarchical Clustering Method : " + method)
            self.__image_processing_utils("Hierarchical_Clustering",
                                  "Hierarchical_Clustering_Method_" + method)

            plt.show()
            plt.close()

    def __visualize_clusters(self, model, output_path, model_name=""):
        """
            Creates visualization of clustering model on given data.
        """
        markers = ["+", "*", "X", "o", "v", "P", "H", "4", "p", "D", "s",
                   "1", "x", "d", "_"]
        colors = ['b', 'g', 'r', 'c', 'm', 'y',
                  '#007BA7', '#ff69b4', '#CD5C5C', '#7eab19', '#1a4572',
                  '#2F4F4F', '#4B0082', '#d11141', '#5b2504']

        # Display ranking on color based on amount data points per cluster
        unique, counts = np.unique(model.labels_, return_counts=True)
        cluster_names = ["Cluster:" + str(cluster_label)
                         for cluster_label in unique]
        self.__display_rank_graph(feature_names=cluster_names,
                                  metric=counts,
                                  title=model_name,
                                  output_path=output_path,
                                  model_name=model_name,
                                  y_title="Clusters",
                                  x_title="Found per cluster")

        self.__vertical_spacing(2)
        pl.figure(figsize=(8, 7))

        # Display clustered graph
        cluster_array = list(range(0, len(cluster_names)))
        scaled_cluster_label = np.hstack(
            (self.__scaled, np.reshape(
                model.labels_.astype(int), (self.__scaled.shape[0], 1))))
        for i in range(0, scaled_cluster_label.shape[0]):
            cluster_label = int(scaled_cluster_label[i][-1])
            cluster_array[cluster_label] = pl.scatter(
                scaled_cluster_label[i, 0], scaled_cluster_label[i, 1],
                c=colors[cluster_label], marker=str(markers[cluster_label]))

        pl.legend(cluster_array, cluster_names)
        pl.title(model_name + ' visualized with data', fontsize=15)
        self.__image_processing_utils(output_path,
                              model_name + "_Visualized_Cluster")
        plt.show()
        plt.close()
        pl.close()

        # Spacing for next model
        self.__vertical_spacing(8)

    def create_kmeans_models(self,
                             n_cluster_list,
                             random_state=None):
        """
            Create kmeans models based on a list of 'n_clusters' values
        """

        if isinstance(n_cluster_list, int):
            n_cluster_list = [n_cluster_list]

        for k_val in n_cluster_list:
            self.__all_cluster_models["Kmeans_Cluster_" + str(k_val)] = KMeans(
                n_clusters=k_val, random_state=random_state).fit(self.__scaled)
            print(
                "Successfully generate Kmeans model on "
                "pre_defined_k={0}".format(k_val))

    def create_kmeans_models_with_elbow_graph(self,
                                              random_state=None):
        """
            Generate models based on the found 'elbow' of the interia values.
        """

        ks = range(1, 15)
        inertias = []

        for k in ks:
            # Create a KMeans instance with k clusters: model
            model = KMeans(n_clusters=k,
                           random_state=random_state).fit(self.__scaled)

            # Append the inertia to the list of inertias
            inertias.append(model.inertia_)

        # Plot ks vs inertias
        plt.figure(figsize=(13, 6))
        plt.plot(ks, inertias, '-o')
        plt.xlabel('Number of clusters, k')
        plt.ylabel('Inertia')
        plt.xticks(ks)
        plt.title("Kmeans Elbow", fontsize=15)

        a = KneeLocator(inertias, ks, curve='convex', direction='decreasing')
        elbow_index = np.where(inertias == a.knee)
        elbow_index = elbow_index[0][0] + 1

        for k_val in [elbow_index - 1, elbow_index, elbow_index + 1]:
            self.__all_cluster_models["Kmeans_Cluster_" + str(k_val)] = KMeans(
                n_clusters=k_val, random_state=random_state).fit(self.__scaled)
            plt.plot(ks[k_val - 1], inertias[k_val - 1], 'r*')

            print(
                "Successfully generate Kmeans model on k_val={0}".format(k_val)
            )

        self.__image_processing_utils("Kmeans",
                              "Kmeans_Visualized_Cluster")
        plt.show()
        plt.close()

    def create_agglomerative_models(self,
                                    n_cluster_list,
                                    linkage_methods=None):
        """
            Create multiple agglomerative models based on a list of
            'n_clusters' values and defined linkage methods.
        """

        if isinstance(n_cluster_list, int):
            n_cluster_list = [n_cluster_list]

        if not linkage_methods:
            linkage_methods = ["ward", "complete", "average", "single"]

        knn_graph = kneighbors_graph(
            self.__scaled, len(
                self.__scaled) - 1, include_self=False)

        for n_clusters in n_cluster_list:
            for connectivity in (None, knn_graph):

                for _, linkage in enumerate(linkage_methods):
                    model = AgglomerativeClustering(linkage=linkage,
                                                    connectivity=connectivity,
                                                    n_clusters=n_clusters)
                    model.fit(self.__scaled)
                    self.__all_cluster_models[
                        "AgglomerativeClustering_{0}_"
                        "cluster{1}_Connectivity{2}".format(
                            linkage,
                            n_clusters, connectivity is not None)] = model

                    print(
                        "Successfully generate Agglomerative model with "
                        "linkage {0} on n_clusters={1}".format(
                            linkage, n_clusters))

    def evaluate_all_models(self,
                            df,
                            df_features,
                            le_map=None,
                            show_extra=True,
                            find_nearest_on_cols=False,
                            zscore_low=-2,
                            zscore_high=2):
        """
            Loop through all models and evaluate the given model with
            'evaluate_model'. Read 'evaluate_model' to learn more.
        """

        for model_name, model in self.__all_cluster_models.items():
            self.evaluate_model(model_name=model_name,
                                model=model,
                                df=df,
                                df_features=df_features,
                                le_map=le_map,
                                find_nearest_on_cols=find_nearest_on_cols,
                                show_extra=show_extra,
                                zscore_low=zscore_low,
                                zscore_high=zscore_high)

            self.__vertical_spacing(5)
            print("----" * 20)

    def evaluate_model(self,
                       model_name,
                       model,
                       df,
                       df_features,
                       output_folder=None,
                       le_map=None,
                       show_extra=True,
                       find_nearest_on_cols=False,
                       zscore_low=-2,
                       zscore_high=2):
        """
        model_name:
            The string key to give the dict

        model:
            Cluster model type; it must have '.labels_' as an attribute

        df:
            Dataframe object

        df_features:
            DataFrameTypes object; organizes feature types into groups.

        output_folder:
            Sub directory to put the pngs

        le_map:
            Dict of dataframe cols to LabelEncoders

        show_extra:
            Show extra information from all functions

        find_nearest_on_cols:
                Allows columns to be converted to actual values found within
                the dataset.
                Ex: Can't have .7 in a bool column convert's it to 1.

                False: Just apply to obj columns and bool columns

                True: Apply to all columns

        zscore_low/zscore_high:
            Defines how the threshold to remove data points when profiling the
            cluster.


        The main purpose of 'evaluate_model' is to display/save tables/plots
        accoiated with describing the model's 'findings'.
        """

        df = copy.deepcopy(df)

        # Create folder structure for png outputs
        if not output_folder:
            output_path = str(model).split("(", 1)[0] + "/" + model_name
        else:
            output_path = output_folder + "/" + model_name

        # ---
        self.__visualize_clusters(model=model,
                                  output_path=output_path,
                                  model_name=model_name)

        # ---
        df["Cluster_Name"] = model.labels_
        numerical_features = df_features.numerical_features()
        clustered_dataframes, shrunken_labeled_df = \
            self.__create_cluster_sub_dfs(
                df=df, model=model, numerical_features=numerical_features,
                zscore_low=zscore_low, zscore_high=zscore_high)

        rows_count, cluster_profiles_df = self.__create_cluster_profiles(
            clustered_dataframes=clustered_dataframes,
            shrunken_df=shrunken_labeled_df,
            numerical_features=df_features.numerical_features(),
            le_map=le_map,
            output_path=output_path,
            show=show_extra,
            find_nearest_on_cols=find_nearest_on_cols)

        # Check to see how many data points were removed during the profiling
        # stage
        print("Orginal points in dataframe: ", df.shape[0])
        print("Total points in all modified clusters: ", rows_count)
        print("Shrank by: ", df.shape[0] - rows_count)

        # In case to many data points were removed
        if cluster_profiles_df.shape[0] == 0:
            print(
                "The applied Z-scores caused the cluster profiles "
                "to shrink to far for the model {0}!".format(
                    model_name))

        # Display and save dataframe table
        else:
            display(cluster_profiles_df)
            self.__render_mpl_table(cluster_profiles_df, sub_dir=output_path,
                                    filename="All_Clusters",
                                    header_columns=0, col_width=2.0)

    # Takes a dict of LabelEncoders to decode the dataframe
    def __decode_df(self, df, le_map):

        df = copy.deepcopy(df)
        decode_cols = list(le_map.keys())
        df[decode_cols] = df[decode_cols].apply(
            lambda x: le_map[x.name].inverse_transform(x))

        return df

    # Not created by me!
    # Author: http://tinyurl.com/y2hjhbwf
    def __render_mpl_table(
            self,
            data,
            sub_dir,
            filename,
            col_width=3.0,
            row_height=0.625,
            font_size=14,
            header_color='#40466e',
            row_colors=['#f1f1f2', 'w'],
            edge_color='w',
            bbox=[0, 0, 1, 1],
            header_columns=0,
            ax=None,
            **kwargs):

        if ax is None:
            size = (np.array(
                data.shape[::-1]) + np.array([0, 1])) * np.array([col_width,
                                                                  row_height])
            fig, ax = plt.subplots(figsize=size)
            ax.axis('off')

        mpl_table = ax.table(
            cellText=data.values,
            bbox=bbox,
            colLabels=data.columns,
            **kwargs)

        mpl_table.auto_set_font_size(False)
        mpl_table.set_fontsize(font_size)

        for k, cell in six.iteritems(mpl_table._cells):
            cell.set_edgecolor(edge_color)
            if k[0] == 0 or k[1] < header_columns:
                cell.set_text_props(weight='bold', color='w')
                cell.set_facecolor(header_color)
            else:
                cell.set_facecolor(row_colors[k[0] % len(row_colors)])

        self.__image_processing_utils(sub_dir, filename)

        plt.close()

    def __create_cluster_profiles(self,
                                  clustered_dataframes,
                                  shrunken_df,
                                  numerical_features,
                                  le_map,
                                  output_path,
                                  find_nearest_on_cols=False,
                                  show=True):
        """
            Profile each clustered dataframe based off the given mean.
            Displays extra information in dataframe tables to be understand
            each cluster.

            find_nearest_on_cols:
                Allows columns to be converted to actual values found within
                the dataset.
                Ex: Can't have .7 in a bool column convert's it to 1.

                False: Just apply to obj columns and bool columns

                True: Apply to all columns
        """

        def find_nearest(numbers, target):
            """
                Find the closest fitting number to the target number
            """
            numbers = np.asarray(numbers)
            idx = (np.abs(numbers - target)).argmin()
            return numbers[idx]

        cluster_profiles_df = pd.DataFrame(columns=shrunken_df.columns).drop(
            'Cluster_Name', axis=1)
        rows_count = 0
        for cluster_identfier, cluster_dataframe in \
                clustered_dataframes.items():
            df = pd.DataFrame(columns=cluster_dataframe.columns)
            df = df.append(cluster_dataframe.mean(), ignore_index=True)
            df.index = [cluster_identfier]

            if cluster_dataframe.shape[0] <= 1:
                continue

            # Attempt to convert numbers found within the full set of data
            for col in cluster_dataframe.columns:
                if col not in numerical_features or find_nearest_on_cols:
                    df[col] = find_nearest(numbers=shrunken_df[
                        col].value_counts().index.tolist(),
                        target=df[col].values[0])

            # Evaluate cluster dataframe by dataframe
            eval_df = pd.DataFrame(columns=cluster_dataframe.columns)
            eval_df = eval_df.append(
                cluster_dataframe.mean(), ignore_index=True)
            eval_df = eval_df.append(
                cluster_dataframe.min(), ignore_index=True)
            eval_df = eval_df.append(
                cluster_dataframe.median(),
                ignore_index=True)
            eval_df = eval_df.append(
                cluster_dataframe.max(), ignore_index=True)
            eval_df = eval_df.append(
                cluster_dataframe.std(), ignore_index=True)
            eval_df = eval_df.append(
                cluster_dataframe.var(), ignore_index=True)
            eval_df.index = ["Mean", "Min", "Median",
                             "Max", "Standard Deviation", "Variance"]

            if show:
                print("Total found in {0} is {1}".format(
                    cluster_identfier, cluster_dataframe.shape[0]))
                self.__render_mpl_table(
                    df,
                    sub_dir=output_path,
                    filename=cluster_identfier +
                    "_Means_Rounded_To_Nearest_Real_Numbers",
                    header_columns=0,
                    col_width=4.0)

                self.__render_mpl_table(
                    eval_df,
                    sub_dir=output_path,
                    filename=cluster_identfier +
                    "_Eval_Df",
                    header_columns=0,
                    col_width=4.0)
                display(df)
                display(eval_df)
                self.__vertical_spacing(7)

            cluster_profiles_df = cluster_profiles_df.append(
                self.__decode_df(df, le_map))

            rows_count += cluster_dataframe.shape[0]

        return rows_count, cluster_profiles_df

    def __create_cluster_sub_dfs(self,
                                 df,
                                 model,
                                 numerical_features,
                                 zscore_low=-2,
                                 zscore_high=2):
        """
            Shrinks the clustered dataframe by rows based on outliers
            found within each cluster.

            Returns back a dict of dataframes with specficed clustered values
            alongside a full dataframe comprised of those clustered dataframes.
        """
        # Dataframe to analyze model 'better' choices
        shrunken_full_df = df.drop('Cluster_Name', axis=1).drop(df.index)

        # Store each sub-dataframe based on cluster label
        clustered_dataframes = dict()

        for cluster_label in set(model.labels_):
            cluster_df = df[df["Cluster_Name"] == cluster_label]
            # Ignore cluster with only one patient
            if len(cluster_df) <= 1:
                continue
            # ---
            zscore_cluster_df = cluster_df.drop(
                'Cluster_Name', axis=1).apply(zscore)

            # Check if cluster is only comprised of one data point
            if cluster_df.shape[0] > 1:

                # Iterate through all numerical features
                for numerical_feature in numerical_features:

                    nan_check = zscore_cluster_df[
                        numerical_feature].isnull().values.any()
                    # Check for nans
                    if not nan_check:
                        zscore_cluster_df = zscore_cluster_df[
                            zscore_cluster_df[numerical_feature] >= zscore_low]
                        zscore_cluster_df = zscore_cluster_df[
                            zscore_cluster_df[numerical_feature] <= zscore_high]

            # Dummy list of -1s alloc at given pos of 'zscore_cluster_df'
            # indexes
            reshaped_index = [-1] * len(df.index.values)

            for given_index in list(zscore_cluster_df.index.values):
                reshaped_index[given_index] = given_index

            # Pass back all vectors that passed the zscore test
            bool_array = pd.Series(reshaped_index).astype(int) == pd.Series(
                list(df.index.values)).astype(int)

            temp_cluster_df = df[bool_array].reset_index(drop=True)

            # Store in proper collection objs
            shrunken_full_df = shrunken_full_df.append(temp_cluster_df)

            clustered_dataframes[
                "Cluster:" + str(cluster_label)] = temp_cluster_df.drop(
                'Cluster_Name', axis=1)

        return clustered_dataframes, shrunken_full_df

    def __inspect_feature_matrix(self,
                                 matrix,
                                 feature_names):
        """
            Creates a dataframe to quickly analyze a matrix
        """
        mean_matrix = np.mean(matrix, axis=0)
        std_matrix = np.std(matrix, axis=0)
        data_dict = dict()
        for index, feature_name in enumerate(feature_names):
            data_dict[feature_name] = [mean_matrix[index],
                                       std_matrix[index]]

        tmp_df = pd.DataFrame.from_dict(data_dict,
                                        orient='index',
                                        columns=['Mean', 'Standard Dev'])

        display(tmp_df)

        return tmp_df

    # Not created by me!
    # Created by my teacher: Narine Hall
    def __visualize_pca_variance(self, data):
        """
            Visualize PCA matrix feature importance
        """

        # Check for pca variance
        pca = PCA()
        data = pca.fit_transform(data)

        # ----
        plt.bar(range(1, len(pca.explained_variance_ratio_) + 1),
                pca.explained_variance_ratio_)
        plt.xticks()
        plt.ylabel('variance ratio')
        plt.xlabel('PCA feature')
        plt.tight_layout()
        self.__image_processing_utils("PCA", "PCA_Feature_Variance_Ratio")
        plt.show()
        plt.close()

        # ----
        plt.bar(range(1, len(pca.explained_variance_ratio_) + 1),
                pca.explained_variance_ratio_.cumsum())
        plt.xticks()
        plt.ylabel('cumulative sum of variances')
        plt.xlabel('PCA feature')
        plt.tight_layout()
        self.__image_processing_utils("PCA", "PCA_Cumulative_Sum_of_Variances")
        plt.show()
        plt.close()

        return pca, data

    # --- Figures maintaining ---
    def __check_create_figure_dir(self,
                                  sub_dir):
        """
            Checks/Creates required directory structures inside
            the parent directory figures.
        """

        directory_path = self.__PROJECT.PATH_TO_OUTPUT_FOLDER

        for dir in sub_dir.split("/"):
            directory_path += "/" + dir
            if not os.path.exists(directory_path):
                os.makedirs(directory_path)

        return directory_path

    def __image_processing_utils(self,
                         sub_dir,
                         filename):
        """
            Saves the plt based image in the correct directory.
        """

        # Ensure directory structure is init correctly
        abs_path = self.__check_create_figure_dir(sub_dir)

        # Ensure file ext is on the file.
        if filename[-4:] != ".png":
            filename += ".png"

        fig = plt.figure(1)
        fig.savefig(abs_path + "/" + filename, bbox_inches='tight')

    # --- Misc
    # I am this lazy yes...
    def __vertical_spacing(self, spaces=1):
        for _ in range(0, spaces):
            print()

    # --- Getters/Setters
    def get_scaled_data(self):
        return copy.deepcopy(self.__scaled)

    def get_all_cluster_models(self):
        return copy.deepcopy(self.__all_cluster_models)

    # def append_model(self,
    #                  model_name,
    #                  model):
    #     try:
    #         model.labels_
    #         self.__all_cluster_models[model_name] = model
    #     except AttributeError:
    #         print("Can not append model to the rest of the models")



from matplotlib import pyplot as plt
from PIL import Image
from PIL import ImageEnhance
import copy

from eflow.utils.sys_utils import create_dir_structure
from eflow.utils.string_utils import correct_directory_path

__author__ = "Eric Cacciavillani"
__copyright__ = "Copyright 2019, eFlow"
__credits__ = ["Eric Cacciavillani"]
__license__ = "MIT"
__maintainer__ = "EricCacciavillani"
__email__ = "eric.cacciavillani@gmail.com"



def create_plt_png(directory_path,
                   sub_dir,
                   filename,
                   sharpness=1.7):

    """

        Saves the plt based image in the correct directory.

    Args:
        directory_path:
            Already existing directory path.

        sub_dir:
            Directory structure to create on top of the already generated path of
            'directory_path'.

        filename:
            Filename to save into the full path of 'directory_path' + 'sub_dir'.

        sharpness:
            Changes the image's sharpness to look better.
    """
    directory_path = correct_directory_path(directory_path)

    # Ensure directory structure is init correctly
    abs_path = create_dir_structure(directory_path,
                                    sub_dir)

    # Ensure file ext is on the file.
    if filename[-4:] != ".png":
        filename += ".png"

    # plt.show()

    plt.savefig(abs_path + "/" + filename, bbox_inches='tight')

    if sharpness:
        full_path = directory_path + sub_dir + "/" + filename
        adjust_sharpness(full_path,
                         full_path,
                         sharpness)

# Taken from: http://tinyurl.com/y6x7nh7t
def adjust_brightness(input_image_path,
                      output_image_path,
                      factor=1.7):
    """

        Adjust brightness of a saved image and save them to a give image path
        with filename.

    Args:
        input_image_path:
            Path to the image file to apply changes.

        output_image:
            Path to output the modified image file.

        factor:
            A float based value to determine the level of effect on the image.
    """

    image = Image.open(input_image_path)
    enhancer_object = ImageEnhance.Brightness(image)
    out = enhancer_object.enhance(factor)
    out.save(output_image_path)

# Taken from: http://tinyurl.com/y6x7nh7t
def adjust_contrast(input_image_path,
                    output_image_path,
                    factor=1.7):
    """

        Adjust contrast of a saved image and save them to a give image path
        with filename.

    Args:
        input_image_path:
            Path to the image file to apply changes.

        output_image:
            Path to output the modified image file.

        factor:
            A float based value to determine the level of effect on the image.
    """
    image = Image.open(input_image_path)
    enhancer_object = ImageEnhance.Contrast(image)
    out = enhancer_object.enhance(factor)
    out.save(output_image_path)

# Taken from: http://tinyurl.com/y6x7nh7t
def adjust_sharpness(input_image_path,
                     output_image_path,
                     factor=1.7):
    """

        Adjust sharpness of a saved image and save them to a give image path
        with filename.

    Args:
        input_image_path:
            Path to the image file to apply changes.

        output_image:
            Path to output the modified image file.

        factor:
            A float based value to determine the level of effect on the image.
    """
    image = Image.open(input_image_path)
    enhancer_object = ImageEnhance.Sharpness(image)
    out = enhancer_object.enhance(factor)
    out.save(output_image_path)


from eflow.utils.string_utils import create_hex_decimal_string

import ipywidgets as widgets
from IPython.display import display
from ipywidgets import Layout

import ipywidgets as widgets
from IPython.display import display
from ipywidgets import Layout

import copy

class ColorLabelingWidget():

    def __init__(self):

        self.__features_w = None
        self.__values_w = None
        self.__color_picker_w = None
        self.__full_widgets_ui = None
        self.__feature_values_w = None

        # ---
        self.__feature_value_color_dict = dict()
        self.__init_features_set = set()

    def get_feature_value_color_dict(self):
        return copy.deepcopy(self.__feature_value_color_dict)

    def run_widget(self,
                   feature_value_color_dict):
        """
        df:
            A pandas dataframe object

        df_features:
            DataFrameTypes object; organizes feature types into groups.

        Returns:
            Returns a UI widget to create a JSON file for cleaning.
        """

        self.__feature_value_color_dict = copy.deepcopy(feature_value_color_dict)

        # Convert all
        for k,v in feature_value_color_dict.items():
            if not v:
                self.__feature_value_color_dict[k] = "#000000"

            if isinstance(v,str):
                del self.__feature_value_color_dict[k]

        feature_values = {feature_name: self.__feature_value_color_dict[feature_name]
                          for feature_name in self.__feature_value_color_dict.keys()}

        self.__features_w = widgets.Select(
            options=list(feature_values.keys()),
            layout=Layout(width='50%',
                          height='175px')
        )

        self.__feature_values_w = {feature: widgets.Select(
             options=feature_values[feature].keys(),
             layout=Layout(width='50%',
                          height='300px'))
             for feature in feature_values}

        self.__color_picker_w = widgets.ColorPicker(
            concise=False,
            description='Pick a color',
            disabled=False
        )

        self.__update_widgets()

        # Link functions with non-updateable widgets
        self.__features_w.observe(self.__select_feature,
                                  'value')

        self.__color_picker_w.observe(self.__select_color,
                                      'value')

        # Setup and display full UI
        self.__full_widgets_ui = widgets.interactive(
            self.__select_value,
            Features=self.__features_w,
            Values=self.__values_w,
            Color_Picker=self.__color_picker_w,
        )

        display(self.__full_widgets_ui)

    def __update_widgets(self):

        init = self.__features_w.value
        self.__values_w = self.__feature_values_w[init]


    def __select_value(self,
                       **func_kwargs):
        """

            Save colors to dictionary and save json file.
        """

        saved_color = self.__feature_value_color_dict[self.__features_w.value][
            self.__values_w.value]

        if not saved_color:
            self.__color_picker_w.value = "#000000"
        else:
            self.__color_picker_w.value = saved_color


    def __select_color(self,
                       _):
        """

            Select the color on color picker widget.
        """

        if self.__color_picker_w.value != "#000000":
            self.__feature_value_color_dict[self.__features_w.value][self.__values_w.value] = self.__color_picker_w.value

    def __select_feature(self,
                         _):
            """

                When a feature selection is chosen all the widgets are
                re-initialized.
            """

            self.__update_widgets()

            new_i = widgets.interactive(self.__select_value,
                                        Features=self.__features_w,
                                        Values=self.__values_w,
                                        Color_Picker=self.__color_picker_w)

            self.__full_widgets_ui.children = new_i.children


from eflow.utils.sys_utils import dict_to_json_file,json_file_to_dict
from eflow.utils.language_processing_utils import get_synonyms
from eflow._hidden.custom_exceptions import UnsatisfiedRequirments
from eflow._hidden.constants import BOOL_STRINGS

import copy

import numpy as np
import pandas as pd
from dateutil import parser
from IPython.display import display
from IPython.display import clear_output

__author__ = "Eric Cacciavillani"
__copyright__ = "Copyright 2019, eFlow"
__credits__ = ["Eric Cacciavillani"]
__license__ = "MIT"
__maintainer__ = "EricCacciavillani"
__email__ = "eric.cacciavillani@gmail.com"


class DataFrameTypes:

    """
        Separates the features based off of dtypes to better keep track of
        feature types and helps make type assertions.
    """

    def __init__(self,
                 df=None,
                 target_feature=None,
                 ignore_nulls=False,
                 fix_numeric_features=False,
                 fix_string_features=False,
                 notebook_mode=False):
        """
        Args:
            df: pd.DataFrame
                Pandas dataframe object.

            target_feature: string
                If the project is using a supervised learning approach we can
                specify the target column. (Note: Not required)

            ignore_nulls: bool
                If set to true than a temporary dataframe is created with each
                feature removes it's nan values to assert what data type the series
                object would be without nans existing inside it.

            fix_numeric_features: bool
                Will attempt to convert all numeric features to the most proper
                numerical types.

            fix_string_features: bool
                Will attempt to convert all string features to ALL proper types.

            notebook_mode: bool
                Boolean value to determine if any notebook functions can be used here.
        """

        # Init an empty dataframe
        if df is None:
            df = pd.DataFrame({})

        # Grab features based on there types
        self.__bool_features = set(
            df.select_dtypes(include=["bool"]).columns)
        self.__string_features = set(
            df.select_dtypes(include=["object"]).columns)
        self.__categorical_features = set(
            df.select_dtypes(include=["category"]).columns)
        self.__integer_features = set(
            df.select_dtypes(include=["int"]).columns)
        self.__float_features = set(
            df.select_dtypes(include=["float"]).columns)
        self.__datetime_features = set(
            df.select_dtypes(include=["datetime"]).columns)

        null_features = df.isnull().sum()
        null_features[null_features == df.shape[0]].index.to_list()
        self.__null_only_features = set(null_features[null_features == df.shape[0]].index.to_list())

        del null_features

        # Target feature for machine learning projects
        self.__target_feature = None

        # Feature's colors
        self.__feature_value_color_dict = dict()

        # Category/Label encoders
        self.__label_encoder = dict()
        self.__label_decoder = dict()

        # Feature values representation
        self.__feature_value_representation = dict()

        # Dummy encoded feature dictionaries
        self.__dummy_encoded_features = dict()

        # Feature's labels and bins
        self.__feature_labels_bins_dict = dict()

        # Data type assertions without nulls
        if ignore_nulls and df.isnull().values.any():
            self.fix_nan_features(df)

        # Attempt to init target column
        if target_feature:
            if target_feature in df.columns:
                self.__target_feature = target_feature
            else:
                raise KeyError(f"The given target feature: \'{target_feature}\' does not exist!")

        # Error checking; flag error; don't disrupt runtime
        features_not_captured = set(df.columns)
        all_features = (self.__float_features | self.__integer_features) | \
                        self.__string_features | \
                        self.__bool_features | \
                        self.__datetime_features | \
                        self.__categorical_features | \
                        self.__null_only_features

        for col_feature in all_features:
            features_not_captured.remove(col_feature)

        if features_not_captured:
            print("ERROR UNKNOWN FEATURE(S) TYPE(S) FOUND!\n{0}".format(
                features_not_captured))

        if fix_string_features:
            self.fix_string_features(df,
                                     notebook_mode)

        if fix_numeric_features:
            self.fix_numeric_features(df,
                                      notebook_mode)

    # --- Getters ---
    def numerical_features(self,
                           exclude_target=False):
        """

            Gets all numerical features chosen by the object.

        Args:
            exclude_target: bool
                If the target feature is an numerical (int/float/bool); then it will be ignored
                when passing back the set.

        Returns:
            Returns a set of all numerical features chosen by the object.
        """
        tmp_set = self.__float_features | self.__integer_features | self.__bool_features
        if exclude_target:

            # Target feature never init
            if self.__target_feature:
                raise KeyError("Target feature was never initialized")

            # Check if target exist in set
            if self.__target_feature in tmp_set:
                tmp_set.remove(self.__target_feature)
            return tmp_set
        else:
            return tmp_set

    def non_numerical_features(self,
                               exclude_target=False):
        """

            Gets all non-numerical features chosen by the object.

        Args:
            exclude_target: bool
                If the target feature is an numerical (int/float/bool); then it will be ignored
                when passing back the set.

        Returns:
            Returns a set of all non numerical features chosen by the object.
        """
        tmp_set = self.all_features() ^ self.numerical_features()
        if exclude_target:

            # Target feature never init
            if self.__target_feature:
                raise KeyError("Target feature was never initialized")

            # Check if target exist in set
            if self.__target_feature in tmp_set:
                tmp_set.remove(self.__target_feature)
            return tmp_set
        else:
            return tmp_set

    def continuous_numerical_features(self,
                                      exclude_target=False):
        """

            Gets all numerical features that are continuous (int/float).

        Args:
            exclude_target: bool
                If the target feature is an numerical (int/float); then it will be ignored
                when passing back the set.

        Returns:
            Returns a set of all numerical features chosen by the object.
        """
        tmp_set = self.__float_features | self.__integer_features
        if exclude_target:

            # Target feature never init
            if self.__target_feature:
                raise KeyError("Target feature was never initialized")

            # Check if target exist in set
            if self.__target_feature in tmp_set:
                tmp_set.remove(self.__target_feature)
            return tmp_set
        else:
            return tmp_set

    def non_continuous_numerical_features(self,
                                          exclude_target=False):
        """

            Gets all numerical features that are not continuous (bool)

        Args:
            exclude_target: bool
                If the target feature is a bool; then it will be ignored
                when passing back the set.

        Returns:
            Returns a set of all numerical features chosen by the object.
        """
        tmp_set = self.__bool_features
        if exclude_target:

            # Target feature never init
            if self.__target_feature:
                raise KeyError("Target feature was never initialized")

            # Check if target exist in set
            if self.__target_feature in tmp_set:
                tmp_set.remove(self.__target_feature)
            return tmp_set
        else:
            return tmp_set

    def continuous_features(self,
                            exclude_target=False):
        """

            Gets all numerical features chosen by the object.

        Args:
            exclude_target: bool
                If the target feature is an numerical (int/float/time); then it will be ignored
                when passing back the set.

        Returns:
            Returns a set of all numerical features chosen by the object.
        """
        tmp_set = self.__float_features | self.__integer_features | self.__datetime_features
        if exclude_target:

            # Target feature never init
            if self.__target_feature:
                raise KeyError("Target feature was never initialized")

            # Check if target exist in set
            if self.__target_feature in tmp_set:
                tmp_set.remove(self.__target_feature)
            return tmp_set
        else:
            return tmp_set

    def non_continuous_features(self,
                                exclude_target=False):
        """

            Gets all numerical features chosen by the object.

        Args:
            exclude_target: bool
                If the target feature is an numerical (int/float/time); then it will be ignored
                when passing back the set.

        Returns:
            Returns a set of all numerical features chosen by the object.
        """
        tmp_set = self.all_features() ^ self.continuous_features()
        if exclude_target:

            # Target feature never init
            if self.__target_feature:
                raise KeyError("Target feature was never initialized")

            # Check if target exist in set
            if self.__target_feature in tmp_set:
                tmp_set.remove(self.__target_feature)
            return tmp_set
        else:
            return tmp_set

    def integer_features(self,
                         exclude_target=False):
        """

            All integer features chosen by df_features.

        Args:
            exclude_target: bool
                If the target feature is an integer; then it will be ignored
                when passing back the set.

        Returns:
            Returns a set of all integer features chosen by the object.
        """
        if exclude_target:
            tmp_set = copy.deepcopy(self.__integer_features)

            # Target feature never init
            if not self.__target_feature:
                raise KeyError("Target feature was never initialized")

            # Check if target exist in set
            if self.__target_feature in tmp_set:
                tmp_set.remove(self.__target_feature)

            return tmp_set
        else:
            return copy.deepcopy(self.__integer_features)

    def float_features(self,
                           exclude_target=False):
        """

            All float features chosen by df_features.

        Args:
            exclude_target: bool
                If the target feature is an float; then it will be ignored
                when passing back the set.

        Returns:
            Returns a set of all float features chosen by the object.
        """
        if exclude_target:
            tmp_set = copy.deepcopy(self.__float_features)

            # Target feature never init
            if not self.__target_feature:
                raise KeyError("Target feature was never initialized")

            # Check if target exist in set
            if self.__target_feature in tmp_set:
                tmp_set.remove(self.__target_feature)

            return tmp_set
        else:
            return copy.deepcopy(self.__float_features)

    def categorical_features(self,
                             exclude_target=False):
        """

            All categorical features chosen by df_features.

        Args:
            exclude_target: bool
                If the target feature is an categorical; then it will be ignored
                when passing back the set.

        Returns:
            Returns a set of all categorical features chosen by the object.
        """
        if exclude_target:
            tmp_set = copy.deepcopy(self.__categorical_features)

            # Target feature never init
            if not self.__target_feature:
                raise KeyError("Target feature was never initialized")

            # Check if target exist in set
            if self.__target_feature in tmp_set:
                tmp_set.remove(self.__target_feature)

            return tmp_set
        else:
            return copy.deepcopy(self.__categorical_features)

    def string_features(self,
                        exclude_target=False):
        """

            All string features chosen by df_features.

        Args:
            exclude_target: bool
                If the target feature is an string; then it will be ignored
                when passing back the set.

        Returns:
            Returns a set of all string features chosen by the object.
        """
        if exclude_target:
            tmp_set = copy.deepcopy(self.__string_features)

            # Target feature never init
            if not self.__target_feature:
                raise KeyError("Target feature was never initialized")

            # Check if target exist in set
            if self.__target_feature in tmp_set:
                tmp_set.remove(self.__target_feature)

            return tmp_set
        else:
            return copy.deepcopy(self.__string_features)

    def bool_features(self,
                      exclude_target=False):

        """

            All bool features chosen by df_features.

        Args:
            exclude_target: bool
                If the target feature is an bool; then it will be ignored
                when passing back the set.

        Returns:
            Returns a set of all bool features chosen by the object.
        """
        if exclude_target:
            tmp_set = copy.deepcopy(self.__bool_features)

            # Target feature never init
            if not self.__target_feature:
                raise KeyError("Target feature was never initialized")

            # Check if target exist in set
            if self.__target_feature in tmp_set:
                tmp_set.remove(self.__target_feature)

            return tmp_set
        else:
            return copy.deepcopy(self.__bool_features)

    def datetime_features(self,
                          exclude_target=False):
        """

            All datetime features chosen by df_features.

        Args:
            exclude_target: bool
                If the target feature is an datetime; then it will be ignored
                when passing back the set.

        Returns:
            Returns a set of all datetime features chosen by the object.
        """
        if exclude_target:
            tmp_set = copy.deepcopy(self.__datetime_features)

            # Target feature never init
            if not self.__target_feature:
                raise KeyError("Target feature was never initialized")

            # Check if target exist in set
            if self.__target_feature in tmp_set:
                tmp_set.remove(self.__target_feature)

            return tmp_set
        else:
            return copy.deepcopy(self.__datetime_features)


    def null_only_features(self,
                           exclude_target=False):


        if exclude_target:
            tmp_set = copy.deepcopy(self.__null_only_features)

            # Target feature never init
            if not self.__target_feature:
                raise KeyError("Target feature was never initialized")

            # Check if target exist in set
            if self.__target_feature in tmp_set:
                tmp_set.remove(self.__target_feature)

            return tmp_set
        else:
            return copy.deepcopy(self.__null_only_features)

    def all_features(self,
                     exclude_target=False):
        """

            Returns all features found in the dataset.

        Args:
            exclude_target: bool
                If the target feature is an datetime; then it will be ignored
                when passing back the set.
        Returns:
            Returns all found features.
        """

        all_features = (self.__float_features | self.__integer_features) | \
                        self.__string_features | \
                        self.__bool_features | \
                        self.__datetime_features | \
                        self.__categorical_features | \
                        self.__null_only_features

        if exclude_target:
            tmp_set = copy.deepcopy(all_features)

            # Target feature never init
            if not self.__target_feature:
                raise KeyError("Target feature was never initialized")

            # Check if target exist in set
            if self.__target_feature in tmp_set:
                tmp_set.remove(self.__target_feature)

            return tmp_set
        else:
            return copy.deepcopy(all_features)


    def get_feature_type(self,
                         feature_name):
        """

            Return's a feature's type as a string

        Args:
            feature_name: str
                The given's feature's name.

        Returns:
            Return's a feature's type as a string
        """

        if feature_name in self.__float_features:
            return "float"

        elif feature_name in self.__bool_features:
            return "bool"

        elif feature_name in self.__integer_features:
            return "integer"

        elif feature_name in self.__categorical_features:
            return "categorical"

        elif feature_name in self.__string_features:
            return "string"

        elif feature_name in self.__datetime_features:
            return "datetime"

        elif feature_name in self.__null_only_features:
            return "null only"

        else:
            raise KeyError(f"Feature '{feature_name}' can't be found in the set!")

    def target_feature(self):
        """

            Gets the target feature.

        Returns:
            Returns the target feature.
        """
        return copy.deepcopy(self.__target_feature)


    def get_label_decoder(self):
        """

            Gets the dict encoder for category to string relationship.

        Returns:
            Returns the encoder dict object.
        """
        return copy.deepcopy(self.__label_decoder)

    def get_label_encoder(self):
        """

            Gets the dict encoder for string to category relationship.

        Returns:
            Returns the encoder dict object.
        """
        return copy.deepcopy(self.__label_encoder)

    def get_feature_colors(self,
                           feature_name):
        """

            Get's the color values for that given feature.

        Args:
            feature_name: str
                The given feature name of the

        Returns:
            Returns the value's color and dictionary; returns None if the feature name
            is not saved in the feature value color dict.
        """
        if feature_name in self.__feature_value_color_dict.keys():
            return copy.deepcopy(self.__feature_value_color_dict[feature_name])
        else:
            return None


    def get_all_feature_colors(self):
        """

           Get's the entire dict of the feature's value colors.

        Returns:
            Returns a copy of feature's value colors.
        """
        return copy.deepcopy(self.__feature_value_color_dict)

    def get_feature_binning(self,
                            feature_name):
        """

           Get's the feature's bin and labels for that given feature.

       Args:
           feature_name: str
               The given feature name of the feature

       Returns:
           Returns the bin and labels for the given feature; returns None if
           the feature name is not saved in the feature value color dict.
       """

        if feature_name in self.__feature_labels_bins_dict:
            return copy.deepcopy(self.__feature_labels_bins_dict[feature_name])
        else:
            return None

    def get_all_feature_binning(self):
        """

           Get's the entire dict of the feature's bins and labels.

        Returns:
            Returns a copy of the labels and bins for the given feature.
        """
        return copy.deepcopy(self.__feature_labels_bins_dict)

    def get_feature_value_representation(self):
        """

            Get's the entire dict of the feature's value colors.

        Returns:
            Returns a copy of feature's value colors.
        """
        return copy.deepcopy(self.__feature_value_representation)

    def get_dummy_encoded_features(self):
        """

            Get's all dummy encoded relationships.
        """
        return copy.deepcopy(self.__dummy_encoded_features)

    # --- Setters and Appending ---
    def set_target_feature(self,
                           feature_name):
        """

            Sets the value.

        Args:
            feature_name:
                Set the target feature.
        """
        self.__target_feature = copy.deepcopy(feature_name)


    def set_feature_to_bool(self,
                            feature_name):
        """

            Moves feature to bool set.

        Args:
            feature_name:
                Feature name or a list of feature names to move to given set.
        """

        if isinstance(feature_name, str):
            feature_name = [feature_name]

        for name in feature_name:
            self.remove_feature(name)
            self.__bool_features.add(name)

    def set_feature_to_integer(self,
                               feature_name):
        """

            Moves feature to integer set.

        Args:
            feature_name:
                Feature name or a list of feature names to move to given set.
        """
        if isinstance(feature_name, str):
            feature_name = [feature_name]

        for name in feature_name:
            self.remove_feature(name)
            self.__integer_features.add(name)

    def set_feature_to_float(self,
                             feature_name):
        """

            Moves feature to float set.

        Args:
            feature_name:
                Feature name or a list of feature names to move to given set.
        """
        if isinstance(feature_name, str):
            feature_name = [feature_name]

        for name in feature_name:
            self.remove_feature(name)
            self.__float_features.add(name)

    def set_feature_to_string(self,
                              feature_name):
        """

            Moves feature to string set.

        Args:
            feature_name:
                Feature name or a list of feature names to move to given set.
        """

        if isinstance(feature_name, str):
            feature_name = [feature_name]

        for name in feature_name:
            self.remove_feature(name)
            self.__string_features.add(name)

    def set_feature_to_categorical(self,
                                   feature_name):
        """

            Moves feature to categorical set.

        Args:
            feature_name:
                Feature name or a list of feature names to move to given set.
        """

        if isinstance(feature_name, str):
            feature_name = [feature_name]

        for name in feature_name:
            self.remove_feature(name)
            self.__categorical_features.add(name)

    def set_feature_to_datetime(self,
                                feature_name):
        """

            Moves feature to datetime set.

        Args:
            feature_name:
                Feature name or a list of feature names to move to given set.
        """

        if isinstance(feature_name,str):
            feature_name = [feature_name]

        for name in feature_name:
            self.remove_feature(name)
            self.__datetime_features.add(name)


    def set_feature_to_null_only_features(self,
                                          feature_name):
        """

            Moves feature to only null series data feature set.

        Args:
            feature_name:
                Feature name or a list of feature names to move to given set.
        """
        if isinstance(feature_name, str):
            feature_name = [feature_name]

        for name in feature_name:
            self.remove_feature(name)
            self.__null_only_features.add(name)

    def set_feature_binning(self,
                            feature_name,
                            bins,
                            labels):

        if not isinstance(labels,list):
            labels = list(labels)

        if not isinstance(bins,list):
            bins = list(bins)

        self.__feature_labels_bins_dict[feature_name] = dict()
        self.__feature_labels_bins_dict[feature_name]["bins"] = bins
        self.__feature_labels_bins_dict[feature_name]["labels"] = labels


    def add_new_bool_feature(self,
                             feature_name):
        """

            Adds a new feature/feature(s) to the feature set bool

        Args:
            feature_name: str
                Name of the new feature
        """

        if isinstance(feature_name,str):
            feature_name = [feature_name]

        for name in feature_name:
            self.__bool_features.add(name)

    def add_new_string_feature(self,
                               feature_name):
        """

            Adds a new feature/feature(s) to the feature set string

        Args:
            feature_name: str
                Name of the new feature
        """
        if isinstance(feature_name, str):
            feature_name = [feature_name]

        for name in feature_name:
            self.__string_features.add(name)

    def add_new_integer_feature(self,
                                feature_name):
        """

            Adds a new feature/feature(s) to the feature set integer

        Args:
            feature_name: str
                Name of the new feature
        """
        if isinstance(feature_name, str):
            feature_name = [feature_name]

        for name in feature_name:
            self.__integer_features.add(name)

    def add_new_float_feature(self,
                              feature_name):
        """

            Adds a new feature/feature(s) to the feature set float

        Args:
            feature_name: str
                Name of the new feature
        """
        if isinstance(feature_name, str):
            feature_name = [feature_name]

        for name in feature_name:
            self.__float_features.add(name)

    def add_new_categorical_feature(self,
                                    feature_name):
        """

            Adds a new feature/feature(s) to the feature set categorical

        Args:
            feature_name: str
                Name of the new feature
        """
        if isinstance(feature_name, str):
            feature_name = [feature_name]

        for name in feature_name:
            self.__categorical_features.add(name)

    def add_new_null_only_feature(self,
                                  feature_name):
        """

            Adds a new feature/feature(s) to the feature set null only features

        Args:
            feature_name: str
                Name of the new feature
        """
        if isinstance(feature_name, str):
            feature_name = [feature_name]

        for name in feature_name:
            self.__null_only_features.add(name)

    def add_new_datetime_feature(self,
                                  feature_name):
        """

            Adds a new feature/feature(s) to the feature set datetime

        Args:
            feature_name: str
                Name of the new feature
        """
        if isinstance(feature_name, str):
            feature_name = [feature_name]

        for name in feature_name:
            self.__datetime_features.add(name)

    def set_feature_colors(self,
                           feature_value_color_dict):
        """

            Passing in a dictionary of feature names to value to hex color to
            save to this object. Error checks the dict for proper values.

        Args:
            feature_value_color_dict:
                Dictionary of feature names to value to hex color.
                Ex: feature_value_color_dict["Sex"]["Male"] = "#ffffff"
        """

        # -----
        for feature_name, color_dict in feature_value_color_dict.items():

            if feature_name not in self.all_features():
                raise UnsatisfiedRequirments(f"The feature name '{feature_name}' "
                                             + "was not found in any of the past"
                                             " feature type sets!")

            if isinstance(feature_name,str):

                # -----
                if isinstance(color_dict,dict):

                    for feature_value, hex_color in color_dict.items():
                        if not isinstance(hex_color, str):
                            raise UnsatisfiedRequirments(f"The feature value must be a string; not a {type(hex_color)}")

                    self.__feature_value_color_dict[feature_name] = color_dict

                # -----
                # elif isinstance(color_dict, str):
                #     try:
                #         sns.color_palette(color_dict)
                #     except:
                #         raise ValueError(f"The value {color_dict} is not a proper seaborn color template!")
                #
                #     self.__feature_value_color_dict[feature_name] = color_dict

                # -----
                else:
                    raise UnsatisfiedRequirments("Expected to extract out a "
                                                 + f"dict from the feature '{feature_name}' "
                                                 + "values with accoiated color hex "
                                                 "values. Instead was found to"
                                                 + f" be a {type(color_dict)}.")
            else:
                raise UnsatisfiedRequirments(f"Expect the feature name to be a "
                                             + f"string instead was found to be {type(feature_name)}")


    def set_feature_value_representation(self,
                                         feature_value_representation):
        for feature_name in feature_value_representation:
            if feature_name not in self.__string_features:
                raise UnsatisfiedRequirments(f"Feature value assertions must be of type string.")

            if feature_name not in self.all_features():
                raise UnsatisfiedRequirments(f"'{feature_name}' doesn't exist in any features.")

        self.__feature_value_representation = copy.deepcopy(feature_value_representation)

    def set_feature_to_dummy_encoded(self,
                                     feature_name,
                                     dummy_encoded_list):
        self.__dummy_encoded_features[feature_name] = dummy_encoded_list

        for bool_feature in dummy_encoded_list:
            self.__bool_features.add(bool_feature)

    # --- Functions ---
    def feature_types_dict(self):
        feature_types = dict()

        # -----
        for feature_name in self.__string_features:
            feature_types[feature_name] = "string"

        for feature_name in self.__bool_features:
            feature_types[feature_name] = "bool"

        for feature_name in self.__integer_features:
            feature_types[feature_name] = "integer"

        for feature_name in self.__float_features:
            feature_types[feature_name] = "float"

        for feature_name in self.__datetime_features:
            feature_types[feature_name] = "datetime"

        for feature_name in self.__categorical_features:
            feature_types[feature_name] = "categorical"

        for feature_name in self.__null_only_features:
            feature_types[feature_name] = "null_only"

        return feature_types

    def feature_types_dataframe(self):
        features = list()
        feature_types = list()

        # -----
        for feature_name in self.__string_features:
            features.append(feature_name)
            feature_types.append("string")

        for feature_name in self.__bool_features:
            features.append(feature_name)
            feature_types.append("bool")

        for feature_name in self.__integer_features:
            features.append(feature_name)
            feature_types.append("integer")

        for feature_name in self.__float_features:
            features.append(feature_name)
            feature_types.append("float")

        for feature_name in self.__datetime_features:
            features.append(feature_name)
            feature_types.append("datetime")

        for feature_name in self.__categorical_features:
            features.append(feature_name)
            feature_types.append("category")

        for feature_name in self.__null_only_features:
            features.append(feature_name)
            feature_types.append("null only")

        dtypes_df = pd.DataFrame({'Data Types': feature_types})
        dtypes_df.index = features
        dtypes_df.index.name = "Features"

        return dtypes_df


    def remove_feature_from_dummy_encoded(self,
                                          feature_name):

        for bool_feature in self.__dummy_encoded_features[feature_name]:
            self.remove_feature(bool_feature)

        del self.__dummy_encoded_features[feature_name]


    def remove_feature(self,
                       feature_name):
        """

            Removes a feature from one of the feature sets.

        Args:
            feature_name:
                The given feature name to remove.
        """
        while True:
            try:
                self.__string_features.remove(feature_name)
                break
            except KeyError:
                pass

            try:
                self.__categorical_features.remove(feature_name)
                break
            except KeyError:
                pass

            try:
                self.__integer_features.remove(feature_name)
                break
            except KeyError:
                pass

            try:
                self.__float_features.remove(feature_name)
                break
            except KeyError:
                pass

            try:
                self.__datetime_features.remove(feature_name)
                break
            except KeyError:
                pass

            try:
                self.__bool_features.remove(feature_name)
                break
            except KeyError:
                pass

            try:
                self.__null_only_features.remove(feature_name)
                break
            except KeyError:
                pass

            raise KeyError(f"The feature {feature_name} doesn't exist inside any of DataFrameType's feature sets!!!")

    def display_features(self,
                         display_dataframes=False,
                         notebook_mode=False):
        """

            Display's the feature sets info.

        Args:
            display_dataframes:
                Creates a dataframe object to display feature information.

            notebook_mode:
                Determines if the dataframe can be displayed in a notebook
        """

        # Do a simple print of feature set info.
        if not display_dataframes:

            # -----
            if self.__string_features:
                print("String Features: {0}\n".format(
                    self.__string_features))

            if self.__categorical_features:
                print("Categorical Features: {0}\n".format(
                    self.__categorical_features))

            if self.__string_features or self.__categorical_features:
                print("---------"*10)

            # -----
            if self.__bool_features:
                print("Bool Features: {0}\n".format(
                    self.__bool_features))
                print("---------" * 10)

            if self.__datetime_features:
                print("Datetime Features: {0}\n".format(
                    self.__datetime_features))
                print("---------" * 10)

            # -----
            if self.__float_features | self.__integer_features:
                print("Numerical Features: {0}\n".format(
                    self.__float_features | self.__integer_features))
            if self.__integer_features:
                print("Integer Features: {0}\n".format(
                    self.__integer_features))

            if self.__float_features:
                print("Float Features: {0}\n".format(
                    self.__float_features))

            if self.__target_feature:
                print("---------" * 10)
                print("Target Feature: {0}\n".format(
                    self.__target_feature))

            if self.__null_only_features:
                print("---------" * 10)
                print("Null Only Feature: {0}\n".format(
                    self.__null_only_features))

        # Create dataframe object based on the feature sets.
        else:
            dtypes_df = self.feature_types_dataframe()

            if notebook_mode:
                display(dtypes_df)
            else:
                print(dtypes_df)

    def fix_numeric_features(self,
                             df,
                             notebook_mode=False,
                             display_results=False):
        """

            Attempts to move numerical features to the correct types by
            following the given priority que:
                1. Bool
                2. Categorical
                3. Float
                4. Int
                5. Do nothing
        Args:
            df: pd.Dataframe
                Pandas Dataframe object to update to correct types.

            notebook_mode: bool
                Boolean value to determine if any notebook functions can be used here.

            display_results: bool
                Display the table in priority order with flags.


        Note -
            This will not actually update the given dataframe. This object is
            a abstract representation of the dataframe.
        """

        features_flag_types = dict()
        for feature_name in df.columns:
            try:
                pd.to_numeric(df[feature_name])
            except ValueError:
                # Ignore all string features
                if feature_name in self.string_features():
                    continue

                # Features that must be these set types
                if feature_name in self.categorical_features():
                    continue

                if feature_name in self.bool_features():
                    continue

            feature_values = set(pd.to_numeric(df[feature_name],
                                 errors="coerce").dropna())

            if not len(feature_values):
                continue

            # Get flag's to push to priority que
            flag_dict = dict()
            flag_dict["Bool"] = self.__bool_check(feature_values)
            numeric_flag, float_flag, int_flag, category_flag = \
                self.__numeric_check(feature_values)
            flag_dict["Numeric"] = numeric_flag
            flag_dict["Float"] = float_flag
            flag_dict["Integer"] = int_flag
            flag_dict["Categorical"] = category_flag

            # Pass the flag dictionary to later be processed by the priority que.
            features_flag_types[feature_name] = flag_dict

        # Iterate on feature and changes based on priority que
        for feature_name, flag_dict in features_flag_types.items():

            # -----
            if flag_dict["Bool"]:
                self.set_feature_to_bool(feature_name)
                continue

            # -----
            elif flag_dict["Categorical"]:
                self.set_feature_to_categorical(feature_name)
                continue

            # -----
            elif flag_dict["Numeric"]:
                if flag_dict["Float"]:
                    self.set_feature_to_float(feature_name)
                    continue

                elif flag_dict["Integer"]:
                    self.set_feature_to_integer(feature_name)
                    continue

        if display_results:
            flag_df = pd.DataFrame.from_dict(features_flag_types,
                                             orient='index')
            if notebook_mode:
                display(flag_df)
            else:
                print(flag_df)

    def fix_string_features(self,
                            df,
                            notebook_mode=False):
        """

            Iterates through all string features and moves features to given types.
            May ask user question if it detects any conflicting string/numeric
            types.

        Args:
            df: pd.Dataframe
                Pandas dataframe object.

            notebook_mode: bool
                Will use the 'clear_output' notebook function if in notebook
                mode.
        """

        # Store types to convert
        type_conflict_dict = dict()

        # Currently this only performs
        for feature_name in self.string_features():

            # Float found
            float_flag = False

            # Keep track of all numeric features
            numeric_count = 0
            numeric_values = []

            # Keep track of all string features
            string_count = 0
            string_values = []

            # Keep track fof all datetime features
            datetime_count = 0
            datetime_values = []

            # Iterate through value counts
            for val, count in df[feature_name].dropna().value_counts().iteritems():

                numeric_check = False

                try:
                    float(val)
                    numeric_check = True
                except ValueError:
                    pass

                # Numeric check
                if isinstance(val, float) or isinstance(val,
                                                        int) or numeric_check == True:
                    numeric_values.append(val)
                    numeric_count += count

                    if isinstance(val, float):
                        float_flag = True

                    if numeric_check and isinstance(val, str):
                        if len(val.split(".")) == 2:
                            float_flag = True

                # String/Datetime check
                elif isinstance(val, str):

                    datetime_found = False
                    try:
                        parser.parse(val)
                        datetime_values.append(val)
                        datetime_count += count
                        datetime_found = True

                    except Exception as e:
                        pass

                    if not datetime_found:
                        string_values.append(val)
                        string_count += count

            # Must be a numeric type; find which type
            if numeric_count != 0 and string_count == 0 and datetime_count == 0:
                if float_flag:
                    type_conflict_dict[feature_name] = "float"
                else:
                    if self.__bool_check(numeric_values):
                        type_conflict_dict[feature_name] = "bool"

                    elif self.__categorical_check(numeric_values):
                        type_conflict_dict[feature_name] = "category"

                    else:
                        type_conflict_dict[feature_name] = "integer"

            # Must be a string type
            elif numeric_count == 0 and string_count != 0 and datetime_count == 0:

                if self.__bool_string_values_check(string_values):
                    type_conflict_dict[feature_name] = "bool"
                else:
                    type_conflict_dict[feature_name] = "string"

            # Must be a datetime
            elif numeric_count == 0 and string_count == 0 and datetime_count != 0:
                type_conflict_dict[feature_name] = "datetime"

            # A conflict is found; have the user work it out.
            else:
                print("Type conflict found!")
                print(f"Feature Name: '{feature_name}'")
                print("---" * 10)

                # -----
                print("Numeric Value Info")
                print(f"\tNumeric count: {numeric_count}")
                print(f"\tNumeric percentage: {(numeric_count / (numeric_count + string_count + datetime_count)) * 100:.3f}%")
                print(f"\tNumeric values: {numeric_values}\n")

                # -----
                print("String Value Info")
                print(f"\tString count: {string_count}")
                print(f"\tString percentage: {(string_count / (numeric_count + string_count + datetime_count)) * 100:.3f}%")
                print(f"\tString values: {string_values}\n")

                # -----
                print("Datetime Value Info")
                print(f"\tString count: {datetime_count}")
                print(f"\tString percentage: {(datetime_count / (numeric_count + string_count + datetime_count)) * 100:.3f}%")
                print(f"\tString values: {datetime_values}\n")

                # Get user input for handling
                print(
                    "You can use the first character of the option for input.\n")
                user_input = input(
                    "\nMove feature to numeric or string and replace any "
                    "conflicts with nulls.\n* Numeric\n* String\n* Datetime\n"
                    "* Ignore\nInput: ")
                user_input = user_input.lower()


                # Clear last user output. If in notebook mode. (A clean notebook is a happy notebook :).)
                if notebook_mode:
                    clear_output()

                # -----

                if not len(user_input):
                    print(f"Ignoring feature '{feature_name}")

                # -----
                elif user_input[0] == "s":
                    type_conflict_dict[feature_name] = "string"

                # -----
                elif user_input[0] == "n":
                    if float_flag:
                        type_conflict_dict[feature_name] = "float"
                    else:
                        numeric_values = set(pd.to_numeric(numeric_values,
                                                       errors="coerce").dropna())
                        if self.__bool_check(numeric_values):
                            type_conflict_dict[feature_name] = "bool"

                        elif self.__categorical_check(numeric_values):
                            type_conflict_dict[feature_name] = "category"

                        else:
                            type_conflict_dict[feature_name] = "integer"

                # -----
                elif user_input[0] == "d":
                    type_conflict_dict[feature_name] = "datetime"

                # -----
                else:
                    print(f"Ignoring feature '{feature_name}")

        # Iterate on all features
        for feature_name, feature_type in type_conflict_dict.items():

            moved_feature = False

            if feature_type == "string":

                if feature_name not in self.__string_features:
                    self.set_feature_to_string(feature_name)
                    moved_feature = True

            elif feature_type == "datetime":

                if feature_name not in self.__datetime_features:
                    self.set_feature_to_datetime(feature_name)
                    moved_feature = True

            elif feature_type == "integer":

                if feature_name not in self.__integer_features:
                    self.set_feature_to_integer(feature_name)
                    moved_feature = True

            elif feature_type == "category":

                if feature_name not in self.__categorical_features:
                    self.set_feature_to_categorical(feature_name)
                    moved_feature = True

            elif feature_type == "bool":

                if feature_name not in self.__bool_features:
                    self.set_feature_to_bool(feature_name)
                    moved_feature = True

            elif feature_type == "float":

                if feature_name not in self.__float_features:
                    self.set_feature_to_float(feature_name)
                    moved_feature = True

            else:
                raise TypeError("An unknown type was passed!")

            if moved_feature:
                print(f"\nMoving feature '{feature_name}' to type {feature_type}.")

    def fix_nan_features(self,
                         df):
        """

            Attempts to get the data type of the feature if there were no nans
            inside it.

        Args:
            df: pd.DataFrame
                Pandas Dataframe object.
        """
        nan_features = [feature for feature, nan_found in
                        df.isna().any().items() if nan_found]

        # The features that are found to be floats should partially merge with
        # features with nulls.
        float_features = set(
            df.select_dtypes(include=["float"]).columns)

        for feature_name in nan_features:

            # Ignore string features
            if feature_name in self.__string_features:
                continue

            feature_values = list(set(df[feature_name].dropna().sort_values(
                ascending=True)))

            # Convert to bool if possible
            if self.__bool_check(feature_values):
                self.set_feature_to_bool(feature_name)

            # Convert numeric to proper types (int,float,categorical)
            elif feature_name in float_features:

                numeric_flag, float_flag, int_flag, category_flag = self.__numeric_check(feature_values)

                if numeric_flag:
                    if category_flag:
                        self.set_feature_to_categorical(feature_name)

                    elif int_flag:
                        self.set_feature_to_integer(feature_name)

                    elif float_flag:
                        self.set_feature_to_float(feature_name)

    def create_json_file_representation(self,
                                        directory_path,
                                        filename):
        """

            Creates a json representation of the current objects feature sets
            and the targeted feature.

        Args:
            directory_path: string
                Absolute directory path.

            filename: string
                File's given name
        """
        type_features = dict()
        for feature_name in self.bool_features():
            if "bool" in type_features.keys():
                type_features["bool"].append(feature_name)
            else:
                type_features["bool"] = [feature_name]

        for feature_name in self.string_features():
            if "string" in type_features.keys():
                type_features["string"].append(feature_name)
            else:
                type_features["string"] = [feature_name]

        for feature_name in self.integer_features():
            if "integer" in type_features.keys():
                type_features["integer"].append(feature_name)
            else:
                type_features["integer"] = [feature_name]

        for feature_name in self.float_features():
            if "float" in type_features.keys():
                type_features["float"].append(feature_name)
            else:
                type_features["float"] = [feature_name]

        for feature_name in self.categorical_features():
            if "categorical" in type_features.keys():
                type_features["categorical"].append(feature_name)
            else:
                type_features["categorical"] = [feature_name]

        for feature_name in self.datetime_features():
            if "datetime" in type_features.keys():
                type_features["datetime"].append(feature_name)
            else:
                type_features["datetime"] = [feature_name]

        type_features["target"] = self.__target_feature

        type_features["feature_labels_bins"] = self.__feature_labels_bins_dict

        type_features["feature_values_colors"] = self.__feature_value_color_dict

        type_features["feature_value_representation"] = self.__feature_value_representation

        type_features["label_encoder"] = self.__label_encoder
        type_features["label_decoder"] = self.__label_decoder

        dict_to_json_file(type_features,
                          directory_path,
                          filename)

    def init_on_json_file(self,
                          filepath):
        """

            Initialize object based on a json file.

        Args:
            filepath:
                Absolute path to the given file.

        Note:
            Structure must be the given type to the list of associated features.
        """
        type_features = json_file_to_dict(filepath)

        # Reset all sets and target
        self.__bool_features = set()
        self.__string_features = set()
        self.__categorical_features = set()
        self.__integer_features = set()
        self.__float_features = set()
        self.__datetime_features = set()

        self.__target_feature = None

        # -----
        self.__feature_value_color_dict = dict()

        # -----
        self.__feature_value_representation = dict()
        self.__label_encoder = dict()
        self.__label_decoder = dict()

        tmp_feature_value_color_dict = dict()

        # Iterate through dict given types to feature lists
        for type, obj  in type_features.items():

            if type == "bool":
                self.__bool_features = set(obj)

            # -----
            elif type == "integer":
                self.__integer_features = set(obj)

            # -----
            elif type == "float":
                self.__float_features = set(obj)

            # -----
            elif type == "string":
                self.__string_features = set(obj)

            # -----
            elif type == "categorical":
                self.__categorical_features = set(obj)

            # -----
            elif type == "datetime":
                self.__datetime_features = set(obj)

            # Extract target
            elif type == "target":
                self.__target_feature = obj

            # Extract color dict.(Naming convention get's a little screwed up here.)
            elif type == "feature_values_colors":
                tmp_feature_value_color_dict = obj

            # Extract out the feature value representation
            elif type == "feature_value_representation":
                self.__feature_value_representation = obj

            elif type == "label_decoder":
                self.__label_decoder = obj

            elif type == "label_encoder":
                self.__label_encoder = obj

            elif type == "feature_labels_bins":
                self.__feature_labels_bins_dict = obj

            else:
                raise ValueError(f"Unknown type {type} was found!")

        # Convert any values that are supposed to numeric back to numeric in colors
        for feature_name,value_color_dict in tmp_feature_value_color_dict.items():
            if feature_name not in self.all_features():
                raise ValueError(f"Unknown feature '{feature_name}' found in color dict for features!")
            else:
                tmp_value_color_dict = copy.deepcopy(value_color_dict)
                for feature_val,color in value_color_dict.items():
                    try:
                        tmp_value_color_dict[int(feature_val)] = color
                        del tmp_value_color_dict[feature_val]
                    except ValueError:
                        pass
                self.__feature_value_color_dict[feature_name] = tmp_value_color_dict

        # Converting to numeric on decoder's categories
        tmp_decoder = self.__label_decoder
        for feature_name,cat_val_dict in tmp_decoder.items():

            cat_val_dict = copy.deepcopy(cat_val_dict)
            for cat, val in cat_val_dict.items():
                self.__label_decoder[feature_name][int(cat)] = self.__label_decoder[feature_name][cat]
                del self.__label_decoder[feature_name][cat]

    def set_encoder_for_features(self,
                                 df,
                                 categorical_value_dict=dict()):
        """

            Create's encoder dict for strings and categories.

        Args:
            df: pd.DataFrame
                Pandas dataframe.

            categorical_value_dict: dict
                Relationship between category and string label.

        Note:
            Can handle a feature that have categories and strings in same series.
        """

        if len(set(df.columns) ^ set(self.all_features())) > 0:
            raise UnsatisfiedRequirments("The given Dataframe's features should "
                                         "be the same as the features saved in "
                                         "the DataFrameTypes.")
        # Reset dict
        self.__label_encoder = dict()

        # Get all string and categorical features
        categorical_string_features = self.categorical_features() | self.string_features()

        for feature_name in categorical_string_features:

            self.__label_encoder[feature_name] = dict()

            # Order feature values
            feature_values = [str(val) for val in
                              df[feature_name].dropna().value_counts(
                                  sort=False).index.to_list()]
            feature_values.sort()

            # Convert to category (int) if possible
            for i in range(0, len(feature_values)):
                try:
                    feature_values[i] = int(float(feature_values[i]))
                except ValueError:
                    pass

            # Pre-defined category values
            default_categories = [i for i in range(0,
                                                   len(feature_values))]

            # Check if user defined any category to string relationships
            if feature_name in categorical_value_dict.keys():
                for cat, label in categorical_value_dict[feature_name].items():

                    # Error checks
                    if label == "":
                        raise UnsatisfiedRequirments(
                            f"Can't replace category value with empty space! Found on feature {feature_name}")

                    if isinstance(label, int):
                        raise UnsatisfiedRequirments(
                            f"Can't replace category value with a number based value! Found on feature {feature_name}")

                    if not isinstance(cat, int):
                        raise UnsatisfiedRequirments(
                            f"Category value must be a number based value! Found on feature {feature_name}")

                    if feature_name not in self.__categorical_features:
                        raise UnsatisfiedRequirments(
                            f"User defined a feature name that doesn't appear to be a '{feature_name}'")

                    self.__label_encoder[feature_name][label] = int(cat)

                    # Remove category and label from collection objects
                    if cat in default_categories:
                        default_categories.remove(cat)

                    if cat in feature_values:
                        feature_values.remove(cat)

                    if label in feature_values:
                        feature_values.remove(label)

            # Encode strings and categories
            cat_count = 0
            for val in feature_values:

                if isinstance(val, int):
                    self.__label_encoder[feature_name][val] = val
                else:
                    self.__label_encoder[feature_name][val] = default_categories[
                        cat_count]
                    cat_count += 1

        # Inverse dict
        self.__label_decoder = dict()
        for feature_name, label_val_dict in self.__label_encoder.items():
            self.__label_decoder[feature_name] = dict()
            for label, cat in label_val_dict.items():
                self.__label_decoder[feature_name][cat] = label

    def __bool_check(self,
                     feature_values):
        """

            Determines if the feature type of this column is a boolean based on
            the numeric values passed to it.

        Args:
            feature_values: collection
                Distinct values of the given feature.

        Returns:
            Returns true or false if the values are boolean.
        """
        # Convert to bool if possible
        if len(feature_values) == 1 and (
                0.0 in feature_values or 1.0 in feature_values):
            return True

        # Second bool check
        elif len(feature_values) == 2 and (
                0.0 in feature_values and 1.0 in feature_values):
            return True
        else:
            return False

    def __bool_string_values_check(self,
                                   feature_values):
        """

            Checks if a collection of strings can be considered a bool feature
            based on the amount of strings and the values of those strings.

        Args:
            feature_values: collection
                Collection object of strings.

        Returns:
            Returns true or false if the values can be considered a bool.
        """

        if len(feature_values) > 2:
            return False

        found_true_value = False
        found_false_value = False

        for val in feature_values:

            if not isinstance(val,str):
                continue

            val = val.lower()

            # Determine if val is true
            if not found_true_value:

                # Check if the string already exist in the defined set
                if val in BOOL_STRINGS.TRUE_STRINGS:
                    found_true_value = True
                    continue
                else:
                    # Attempt to find synonyms of the defined words to compare to
                    # the iterable string
                    for true_string in BOOL_STRINGS.TRUE_STRINGS:

                        if len(true_string) < 2:
                            continue

                        for syn in get_synonyms(true_string):
                            if syn == val:
                                found_true_value = True
                                continue

            # -----
            if not found_false_value:

                # -----
                if val in BOOL_STRINGS.FALSE_STRINGS:
                    found_false_value = True
                    continue
                else:
                    # -----
                    for false_string in BOOL_STRINGS.FALSE_STRINGS:

                        if len(false_string) < 2:
                            continue

                        for syn in get_synonyms(false_string):
                            if syn == val:
                                found_false_value = True
                                continue

        if len(feature_values) == 2:
            return found_true_value and found_false_value
        else:
            return found_true_value or found_false_value

    def __categorical_check(self,
                            feature_values):
        """

            Check to see if the feature's value can be consider a category.

        Args:
            feature_values:
                Distinct values of the given feature.

        Returns:
            Returns true or false if the values can be categorical.
        """

        feature_values = copy.deepcopy(set(feature_values))

        min_val = min(feature_values)
        max_val = max(feature_values)

        if sum(feature_values) != ((min_val + max_val) / 2) * len(
                set(feature_values)):
            return False
        else:
            return True

    def __numeric_check(self,
                        feature_values):
        """

            Checks if the features can be a numerical value and other numerical
            types like floats, ints, and categories.

        Args:
            feature_values:
                Distinct values of the given feature.

        Returns:
            Boolean values return as such:
                numerical_check, float_check, int_check, categorical_check
        """

        if not len(feature_values):
            return False, False, False, False


        # -----
        float_check = False
        categorical_total_sum = 0
        categorical_check = True

        feature_values = copy.deepcopy(set(feature_values))

        # Iterate through all values
        for val in feature_values:

            # Ignore None values
            if not val:
                continue
            try:
                float(val)
            except ValueError:
                return False, False, False, False

            if categorical_check:
                # Check if float
                str_val = str(val)
                tokens = str_val.split(".")
                if len(tokens) > 1 and int(tokens[1]) > 0:
                    float_check = True
                    categorical_check = False
                else:
                    categorical_total_sum += val
                    continue

        # Check if categorical
        if categorical_check:

            min_val = min(feature_values)
            max_val = max(feature_values)

            if categorical_total_sum != ((min_val + max_val)/2) * len(feature_values):
                categorical_check = False

        return True, float_check, not float_check, categorical_check


    def __eq__(self,
               other):

        if isinstance(other,DataFrameTypes):

            if self.feature_types_dict() != other.feature_types_dict():
                return False

            elif self.get_label_encoder() != other.get_label_encoder():
                return False

            elif self.get_label_decoder() != other.get_label_decoder():
                return False

            elif self.get_all_feature_colors() != other.get_all_feature_colors():
                return False

            elif self.get_all_feature_binning() != other.get_all_feature_binning():
                return False

            elif self.target_feature() != other.target_feature():
                return False

            else:
                return True

        else:
            return False